﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace Kokomo.Mirror.Ia32
{
	public class IaReader
	{
		public IaReader(
			Stream stream,
			long streamBaseAddress,
			OperandSize defaultRegisterSize,
			IaOpCodeLookup opCodeLookup
			)
		{
			if (stream == null) throw new ArgumentNullException("stream");
			if (opCodeLookup == null) throw new ArgumentNullException("opCodeLookup");

			this.Stream = stream;
			this.StreamBaseAddress = streamBaseAddress;
			this.Reader = new BinaryReader(stream);
			this.DefaultRegisterSize = defaultRegisterSize;
			this._opCodeLookup = opCodeLookup;
		}

		private IaOpCodeLookup _opCodeLookup;

		public long StreamBaseAddress { get; private set; }
		public OperandSize DefaultRegisterSize { get; private set; }
		public Stream Stream { get; private set; }
		protected BinaryReader Reader { get; private set; }
		public Ia32Instruction CurrentInstruction { get; private set; }

		public bool MoveNext()
		{
			long startStreamOffset = this.Stream.Position;
			long startOffset = startStreamOffset + this.StreamBaseAddress;

			Prefixes prefixes = 0;
			Register segmentOverrideRegister = null;

			int opCodeSize = 1;
			int opCodeValue = 0;
			while (true)
			{
				int n = this.Stream.ReadByte();
				if (n == -1) return false;

				opCodeValue |= n;
				if (n == IaOpCodeLookup.EscapeCode)
				{
					opCodeSize = 2;
					opCodeValue <<= 8;
					continue;
				}

				if (this._opCodeLookup.IsPrefix(opCodeValue))
				{
					prefixes |= this._opCodeLookup.GetPrefixCode(opCodeValue);
					opCodeValue = 0;

					switch (prefixes)
					{
						case Prefixes.CsSegmentOverride:
							segmentOverrideRegister = Registers.CS;
							break;
						case Prefixes.SsSegmentOverride:
							segmentOverrideRegister = Registers.SS;
							break;
						case Prefixes.DsSegmentOverride:
							segmentOverrideRegister = Registers.DS;
							break;
						case Prefixes.EsSegmentOverride:
							segmentOverrideRegister = Registers.ES;
							break;
						case Prefixes.FsSegmentOverride:
							segmentOverrideRegister = Registers.FS;
							break;
						case Prefixes.GsSegmentOverride:
							segmentOverrideRegister = Registers.GS;
							break;
					}

					continue;
				}

				bool hasModRM = false;
				ModRM modrm = new ModRM();
				OpCode opCode;
				if (this._opCodeLookup.IsExtendedOpCode(opCodeValue))
				{
					modrm = new ModRM(this.Reader.ReadByte());
					hasModRM = true;

					opCode = this._opCodeLookup.GetExtendedOpCode(opCodeValue, modrm, prefixes);
				}
				else
				{
					opCode = this._opCodeLookup.GetMatchingOpCode(opCodeValue, opCodeSize);
				}

				if (opCode.IsUnknown) return false;

				ScaleIndexBase sib = new ScaleIndexBase();
				if (opCode.Form.HasFlag(OpCodeForm.ModRM) && !hasModRM)
				{
					modrm = new ModRM((byte)this.Stream.ReadByte());
					hasModRM = true;
				}

				int registerNumber = 0;
				int displacementSizeBytes = 0;
				if (hasModRM)
				{
					registerNumber = modrm.Reg;
					if (prefixes.HasFlag(Prefixes.Rex_r)) registerNumber |= 0x8;

					Mod mod = modrm.Mod;
					if (mod == Mod.disp8) displacementSizeBytes = 1;
					else if (mod == Mod.disp32 || modrm.IndicatesDisp32)
						displacementSizeBytes = 4;

					if (modrm.IndicatesSib)
					{
						sib = new ScaleIndexBase((byte)this.Stream.ReadByte());
						if (
							(sib.Base == Base.disp)
							&& (displacementSizeBytes == 0)
							)
						{
							displacementSizeBytes = 4;
						}
					}
				}

				long signedDisplacement = 0;
				ulong unsignedDisplacement = 0;
				switch (displacementSizeBytes)
				{
					case 0:
						break;
					case 1:
						{
							sbyte b = this.Reader.ReadSByte();
							signedDisplacement = b;
							unsignedDisplacement = (byte)b;
						}
						break;
					case 4:
						{
							int i = this.Reader.ReadInt32();
							signedDisplacement = i;
							unsignedDisplacement = (uint)i;
						}
						break;
					case 8:
						{
							long l = this.Reader.ReadInt64();
							signedDisplacement = l;
							unsignedDisplacement = (ulong)l;
						}
						break;
					default:
						throw new InvalidOperationException("Unknown displacement");
				}

				Operand[] operands = new Operand[opCode.OperandTypes.Length];
				for (int i = 0; i < operands.Length; i++)
				{
					OperandType operandType = opCode.OperandTypes[i];

					OperandType operandEncoding = operandType & OperandType.EncodingMask;
					OperandType operandDataType = operandType & OperandType.DataTypeMask;
					OperandType operandSemantic = operandType & OperandType.SemanticMask;

					OperandSize size = this.DetermineOperandSize(operandDataType, prefixes);

					Operand operand;
					switch (operandEncoding)
					{
						case OperandType.OpCodeReg:
							{
								int opCodeRegisterNumber = n & 0x07;
								if (prefixes.HasFlag(Prefixes.Rex_b)) opCodeRegisterNumber |= 0x8;
								RegisterType registerType = (RegisterType)((int)operandSemantic >> (int)OperandType.RegisterTypeShift);
								Register register = Register.Get(opCodeRegisterNumber, size, registerType);
								if (register == null) throw new InvalidOperationException("Unrecognized register");
								operand = register;
							}
							break;
						case OperandType.ModRM:
							if (modrm.Mod == Mod.reg)
							{
								RegisterType registerType;
								switch (operandSemantic)
								{
									case OperandType.GeneralPurposeRegister:
									case OperandType.GprOrAddress:
										registerType = RegisterType.GeneralPurpose;
										break;
									case OperandType.Mmx:
									case OperandType.MmxOrAddress:
										registerType = RegisterType.Mmx;
										break;
									case OperandType.Xmm:
									case OperandType.XmmOrAddress:
										registerType = RegisterType.Simd;
										break;
									case OperandType.FpuRegister:
									case OperandType.FpuOrAddress:
										registerType = RegisterType.Fpu;
										break;
									case OperandType.ControlRegister:
										registerType = RegisterType.Control;
										break;
									default:
										throw new InvalidOperationException("Unrecognized operand");
								}

								int registerCode = modrm.RM;
								if (prefixes.HasFlag(Prefixes.Rex_b)) registerCode |= 0x8;
								Register register = Register.Get(registerCode, size, registerType);
								if (register == null) throw new InvalidOperationException("Unrecognized operand");

								operand = register;
							}
							else
							{
								if (modrm.IndicatesSib)
								{
									int baseCode = (int)sib.Base;
									if (prefixes.HasFlag(Prefixes.Rex_b)) baseCode |= 0x8;
									SimpleOperand baseOperand;
									if (baseCode == (int)Base.disp)
									{
										baseOperand = new ImmediateOperand(size, unsignedDisplacement);
										signedDisplacement = 0;
										unsignedDisplacement = 0;
									}
									else
									{
										baseOperand = Register.Get(baseCode, this.DefaultRegisterSize, RegisterType.GeneralPurpose);
									}

									int indexCode = (int)sib.Index;
									if (prefixes.HasFlag(Prefixes.Rex_x)) indexCode |= 0x8;
									SimpleOperand indexOperand = null;
									if (indexCode != (int)Index.None)
										indexOperand = Register.Get(indexCode, this.DefaultRegisterSize, RegisterType.GeneralPurpose);

									IndexedOperand indexed = new IndexedOperand(
										size,
										baseOperand,
										indexOperand,
										sib.ScaleFactor,
										signedDisplacement,
										segmentOverrideRegister
										);
									operand = indexed;
								}
								else
								{
									IndirectOperand indirect;
									if (modrm.IndicatesDisp32)
									{
										indirect = new IndirectOperand(size, new ImmediateOperand(size, unsignedDisplacement));
									}
									else
									{
										int baseCode = (int)sib.Base;
										if (prefixes.HasFlag(Prefixes.Rex_b)) baseCode |= 0x8;
										SimpleOperand baseOperand = Register.Get(baseCode, this.DefaultRegisterSize, RegisterType.GeneralPurpose);
										indirect = new IndirectOperand(size, baseOperand, signedDisplacement);
									}

									operand = indirect;
								}
							}
							break;
						case OperandType.Reg:
							{
								Register register = Register.Get(registerNumber, size, (RegisterType)((int)operandSemantic >> (int)OperandType.RegisterTypeShift));
								if (register == null) throw new InvalidOperationException("Unrecognized register");
								operand = register;
							}
							break;
						case OperandType.FixedRegister:
							{
								int registerCode = ((int)operandType & (int)OperandType.FixedRegisterMask) >> (int)OperandType.FixedRegisterShift;
								RegisterType registerType = (RegisterType)((int)operandSemantic >> (int)OperandType.RegisterTypeShift);
								Register register = Register.Get(registerCode, size, registerType);
								if (register == null) throw new InvalidOperationException("Unrecognized register");
								operand = register;
							}
							break;
						case OperandType.ImplicitPointerDI:
							operand = new IndirectOperand(size, Registers.EDI);
							break;
						case OperandType.ImplicitPointerSI:
							operand = new IndirectOperand(size, Registers.ESI);
							break;
						case OperandType.Immediate:
							{
								switch (operandSemantic)
								{
									case OperandType.JumpRelativeOffset:
										int relativeOffset;
										switch (size)
										{
											case OperandSize.Byte:
												relativeOffset = this.Reader.ReadSByte();
												break;
											case OperandSize.Word:
												relativeOffset = this.Reader.ReadInt16();
												break;
											case OperandSize.Dword:
												relativeOffset = this.Reader.ReadInt32();
												break;
											default:
												throw new InvalidOperationException("Unrecognized immediate size");
										}

										operand = new JumpTarget(startOffset, Stream.Position + this.StreamBaseAddress + relativeOffset);
										break;
									case OperandType.JumpAbsoluteOffset:
									case OperandType.Address:
										uint address;
										int addressSize = this.DetermineAddressSize(prefixes);
										switch (addressSize)
										{
											case 1:
												address = this.Reader.ReadByte();
												break;
											case 2:
												address = this.Reader.ReadUInt16();
												break;
											case 4:
												address = this.Reader.ReadUInt32();
												break;
											default:
												throw new InvalidOperationException("Unrecognized immediate size");
										}

										if (operandSemantic == OperandType.JumpAbsoluteOffset)
											operand = new JumpTarget(startOffset, address);
										else
											operand = new IndirectOperand(size, new ImmediateOperand(size, address), 0, segmentOverrideRegister);

										break;
									default:
										ulong immediateValue;
										switch (size)
										{
											case OperandSize.Byte:
												immediateValue = this.Reader.ReadByte();
												break;
											case OperandSize.Word:
												immediateValue = this.Reader.ReadUInt16();
												break;
											case OperandSize.Dword:
												immediateValue = this.Reader.ReadUInt32();
												break;
											case OperandSize.Qword:
												immediateValue = this.Reader.ReadUInt64();
												break;
											default:
												throw new InvalidOperationException("Unrecognized immediate size");
										}

										ImmediateOperand immediate = new ImmediateOperand(size, immediateValue);
										operand = immediate;
										break;
								}
							}
							break;
						default:
							throw new InvalidOperationException("Unrecognized operand type");
							break;
					}

					operands[i] = operand;
				}

				long endStreamOffset = this.Stream.Position;
				int instructionLength = (int)(endStreamOffset - startStreamOffset);
				byte[] bytes = new byte[instructionLength];
				this.Stream.Position = startStreamOffset;
				this.Stream.Read(bytes, 0, instructionLength);
				Ia32Instruction instruction = new Ia32Instruction(
					bytes,
					startOffset,
					instructionLength,
					prefixes,
					opCode,
					operands
					);

				this.CurrentInstruction = instruction;

				return true;
			}
		}

		private int DetermineAddressSize(
			Prefixes prefixes
			)
		{
			int size = prefixes.HasFlag(Prefixes.AddressSizeOverride)
				? 2
				: 4
				;
			return size;
		}

		private OperandSize DetermineOperandSize(
			OperandType sizeType,
			Prefixes prefixes
			)
		{
			OperandSize size;

			bool hasOperandSizeOverridePrefix = prefixes.HasFlag(Prefixes.OperandSizeOverride);
			bool hasAddressSizeOverridePrefix = prefixes.HasFlag(Prefixes.AddressSizeOverride);

			bool hasRexW = prefixes.HasFlag(Prefixes.Rex_w);

			switch (sizeType)
			{
				case OperandType.Byte:
					size = OperandSize.Byte;
					break;
				case OperandType.Word:
					size = OperandSize.Word;
					break;
				case OperandType.Dword:
					size = OperandSize.Dword;
					break;
				case OperandType.Qword:
				case OperandType.DDword:
					size = OperandSize.Qword;
					break;
				case OperandType.DQWord:
					size = OperandSize.DQword;
					break;
				case OperandType.PackedSingle:
					size = OperandSize.DQword;
					break;
				case OperandType.OperandByteOrWord:
					size = (hasOperandSizeOverridePrefix ? OperandSize.Byte : OperandSize.Word);
					break;
				case OperandType.OperandWordOrDword:
					size = (hasOperandSizeOverridePrefix ? OperandSize.Word : OperandSize.Dword);
					break;
				case OperandType.OperandWordDwordQword:
					size = (hasRexW ? OperandSize.Qword : (hasOperandSizeOverridePrefix ? OperandSize.Word : OperandSize.Dword));
					break;
				case OperandType.OperandDwordOrFword:
					size = OperandSize.Dword;
					break;
				case OperandType.AddressWordOrDword:
					size = (hasAddressSizeOverridePrefix ? OperandSize.Word : OperandSize.Dword);
					break;
				case OperandType.DoubleFloatingPoint:
					size = OperandSize.Qword;
					break;
				case OperandType.PseudoDescriptor:
					size = OperandSize.Bytes6;
					break;
				default:
					throw new ArgumentException("Invalid operand size");
			}

			return size;
		}
	}
}
